import abc
import logging
import uuid
import base64
import struct
import binascii
import functools
import dataclasses

from operator import itemgetter

from typing import Union

from . import exceptions

logger = logging.getLogger(__name__)


class ProtectedString(str):
    pass


class LdapString(str):
    def __new__(cls, string, _):
        return super().__new__(cls, string)  # type: ignore

    def __init__(self, string, ldap_attribute):
        super().__init__()
        self.string = string
        self.ldap_attribute = ldap_attribute

    def __repr__(self):
        return f"{self.__class__.__name__}({self.ldap_attribute}='{self.__str__()}')"


class EncodedQueryString(LdapString):
    """
    A type of `str` for holding encoded LDAP query values.

    Instances of this class behave exactly like normal `str` objects.

    Each instance has an `ldap_attribute` attribute. You can use this should you need to find the name of the LDAP
    attribute the value is associated with.
    """


class DecodedLdapString(LdapString):
    """
    A type of `str` for holding decoded LDAP attribute values.

    Instances of this class behave exactly like normal `str` objects.

    Each instance has an `ldap_attribute` attribute. You can use this should you need to find the name of the LDAP
    attribute the value is associated with.
    """


class EncodedLdapBytes(bytes):
    """
    A type of `bytes` for holding encoded LDAP query values.

    Instances of this class behave exactly like normal `bytes` objects.

    Each instance has an `ldap_attribute` attribute. You can use this should you need to find the name of the LDAP
    attribute the value is associated with.
    """

    def __new__(cls, data, _):
        return super().__new__(cls, data)  # type: ignore

    def __init__(self, data: bytes, ldap_attribute: str):
        super().__init__()
        self.bytes = data
        self.ldap_attribute = ldap_attribute

    def __repr__(self):
        return f"{self.__class__.__name__}({self.ldap_attribute}={self.__str__()})"


class Codec(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def decode(cls, data):
        """ A mandatory codec method for decoding an LDAP attribute to a human readable string.
        """

    @classmethod
    @abc.abstractmethod
    def bytes_encode(cls, data):
        """ A mandatory codec method for bytes-encoding an LDAP attribute back to its original value.
        """

    @classmethod
    @abc.abstractmethod
    def query_encode(cls, data):
        """ A mandatory codec method for query-encoding an LDAP attribute value.
        """


class ThumbnailPhotoCodec(Codec):
    name = "thumbnailPhoto"

    @classmethod
    def decode(cls, data: Union[EncodedLdapBytes, bytes]) -> DecodedLdapString:
        if not isinstance(data, (EncodedLdapBytes, bytes)):
            raise exceptions.AttributeDecodingError(f"Can only decode {cls.name} from bytes. Received {type(data)}.")
        logger.debug(f"decoding {cls.name} from bytes value {data}")
        return DecodedLdapString(base64.b64encode(data).decode("utf-8"), cls.name)

    @classmethod
    def bytes_encode(cls, data: Union[DecodedLdapString, str]) -> EncodedLdapBytes:
        if not isinstance(data, (DecodedLdapString, str)):
            raise exceptions.AttributeDecodingError(f"Can only encode {cls.name} from string. Received {type(data)}.")
        logger.debug(f"encoding {cls.name} from string value {data}")
        return EncodedLdapBytes(base64.b64decode(bytes(data, "utf-8")), cls.name)

    @classmethod
    def query_encode(cls, data):
        raise exceptions.UnsupportedCodecFunction(f"Query encoding the {cls.name} attribute is not supported.")


class LogonHoursCodec(Codec):
    name = "logonHours"

    @classmethod
    def decode(cls, data: bytes) -> DecodedLdapString:
        # A byte array of 21 bytes, each represented by two hexadecimal characters escaped with the backslash escape
        # character. Each byte is 8 bits, so the 21 bytes represent the 168 hours in a week (21 x 8 =  7 x 24 = 168).
        #
        # In: b'\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
        # Out: [255, 255, 255, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255]
        if not isinstance(data, bytes):
            raise exceptions.AttributeDecodingError(f"Can only decode {cls.name} from bytes. Received {type(data)}.")
        logger.debug(f"decoding {cls.name} from bytes value {data}")
        return DecodedLdapString(
            " ".join([str(struct.unpack("B", data[index : index + 1])[0]) for index in range(0, len(data))]), cls.name
        )

    @classmethod
    def bytes_encode(cls, data: str) -> EncodedLdapBytes:
        # In: [255, 255, 255, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255]
        # Out: b'\xff\xff\xff\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
        if not isinstance(data, list):
            raise exceptions.AttributeEncodingError(f"Can only encode {cls.name} from a list. Received {type(data)}.")
        logger.debug(f"encoding {cls.name} from string value {data}")
        return EncodedLdapBytes(b"".join([struct.pack("B", value) for value in data]), cls.name)

    @classmethod
    def query_encode(cls, data):
        raise exceptions.UnsupportedCodecFunction(f"Query encoding the {cls.name} attribute is not supported.")


class ObjectGUIDCodec(Codec):
    name = "objectGUID"

    @classmethod
    def decode(cls, data: Union[bytes, EncodedLdapBytes]) -> DecodedLdapString:
        if not isinstance(data, (bytes, EncodedLdapBytes)):
            raise exceptions.AttributeDecodingError(f"Can only decode {cls.name} from bytes. Received {type(data)}.")
        logger.debug(f"decoding {cls.name} from bytes value {data}")
        try:
            return DecodedLdapString(str(uuid.UUID(bytes_le=data)), cls.name)
        except Exception as e:
            logger.exception(e)
            raise exceptions.AttributeDecodingError(f"Failed to decode {cls.name} from value {data}.")

    @classmethod
    def bytes_encode(cls, data: Union[str, DecodedLdapString]) -> EncodedLdapBytes:
        if not isinstance(data, (DecodedLdapString, str)):
            raise exceptions.AttributeEncodingError(
                f"Can only encode {cls.name} from human readable GUID string. Received {type(data)}."
            )
        logger.debug(f"encoding {cls.name} from string value {data}")
        return EncodedLdapBytes(uuid.UUID(data).bytes_le, cls.name)

    @classmethod
    def query_encode(cls, data: Union[DecodedLdapString, EncodedLdapBytes, str, bytes]) -> EncodedQueryString:
        @functools.singledispatch
        def _cast(data):
            return data

        @_cast.register
        def _bytes(data: bytes):
            return cls.decode(data)

        @_cast.register
        def _encoded_bytes(data: EncodedLdapBytes):
            return cls.decode(data)

        decoded_string = _cast(data)

        logger.debug(f"query encoding {cls.name} from string value {data}")

        def _escape_bytes(data):
            # Turn the raw bytes value into a bytes hex representation, then decode it to a utf-8 string.
            # The resulting string will be twice as long as the original data as each hex representation
            # will be 2 digits (hence range being called with steps equal to 2).
            #
            # IN: b'9\x1b]\xef03\x8dD\x89\xc8A\xeac_\xef*'
            # OUT: '\39\1b\5d\ef\30\33\8d\44\89\c8\41\ea\63\5f\ef\2a'
            hex_string = binascii.hexlify(data).decode("utf-8")
            return "".join([f"\\{hex_string[i: i + 2]}" for i in range(0, len(hex_string), 2)])

        def _guid_byte_swap(guid_bytes):
            """ Swaps Guid bytes

            LDAP stores the GUID bytes in a different order than returned by
            uuid.bytes. This function does the conversion in both directions.
            """
            mapper = itemgetter(3, 2, 1, 0, 5, 4, 7, 6, 8, 9, 10, 11, 12, 13, 14, 15)
            return bytes(mapper(guid_bytes))

        try:
            return EncodedQueryString(_escape_bytes(_guid_byte_swap(uuid.UUID(decoded_string).bytes)), cls.name)
        except (ValueError, AttributeError):
            raise exceptions.AttributeEncodingError(
                f"Cannot query-encode {cls.name} from data {data} of type {type(data)}."
            )


class ObjectSidCodec(Codec):
    name = "objectSid"

    @classmethod
    def decode(cls, data: bytes) -> DecodedLdapString:
        if not isinstance(data, bytes):
            raise exceptions.AttributeDecodingError(f"Can only decode {cls.name} from bytes. Received {type(data)}.")
        # In: b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x10;\x01\x00'
        # Out: 'S-1-5-21-1646645597-476216284-2431080400-80656'
        logger.debug(f"decoding {cls.name} from bytes value {data}")

        def _little_endian(chunk):
            return struct.unpack("<q", chunk.ljust(8, b"\x00"))[0]

        def _big_endian(chunk):
            return struct.unpack(">q", chunk.rjust(8, b"\x00"))[0]

        def _unpack(data):
            output = ["S"]
            output.append(_little_endian(data[0:1]))
            output.append(_big_endian(data[2 : 2 + 6]))
            for i in range(8, len(data), 4):
                output.append(_little_endian(data[i : i + 4]))
            logger.debug(f"decoded SID chunks {output}")
            return "-".join(map(str, output))

        return DecodedLdapString(_unpack(data), cls.name)

    @classmethod
    def bytes_encode(cls, data: Union[str, DecodedLdapString]) -> EncodedLdapBytes:
        # In: 'S-1-5-21-1646645597-476216284-2431080400-80656'
        # Out: b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x10;\x01\x00'
        if not isinstance(data, (DecodedLdapString, str)):
            raise exceptions.AttributeEncodingError(
                f"Can only encode {cls.name} from human readable string. Received {type(data)}."
            )
        logger.debug(f"encoding {cls.name} from string value {data}")
        try:
            chunks = [int(i) for i in data[2:].split("-")]
        except ValueError as e:
            logger.exception(e)
            raise exceptions.AttributeEncodingError(f"Failed to encode string {cls.name} {repr(data)} to bytes.")
        encoded_chunks = []

        # Section 1
        #
        # SID Revision
        #
        # Format: little-endian 8-bit unsigned int
        # Human-Readable Section: 'S-1-5-21-1646645597-476216284-2431080400-80656'
        #                            ^
        # Bytes Section: \x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x10;\x01\x00
        #                ^^^^
        #
        encoded_chunks.append(struct.pack("<i", chunks.pop(0)).rstrip(b"\x00"))

        # Section 2
        #
        # Identifier Authority
        #
        # Format: little-endian 8-bit unsigned int
        # Human-Readable Section: 'S-1-5-21-1646645597-476216284-2431080400-80656'
        #                              ^
        # Bytes Section: \x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x10;\x01\x00
        #                    ^^^^
        encoded_chunks.append(struct.pack("<i", chunks.pop(0)).rstrip(b"\x00"))

        # Section 3
        #
        # Sub-authority Count
        #
        # Identifies how many more sub-authority values there are in the SID. This figure
        # is just the length of the remaining human-readable chunks.
        #
        # Format: 6 byte big-endian unsigned long
        # Human-Readable Section: ** Not Displayed - it is metadata **
        # Bytes Section: \x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x10;\x01\x00
        #                        ^^^^^^^^^^^^^^^^^^^^^^^^
        encoded_chunks.append(struct.pack(">L", len(chunks)).rjust(6, b"\x00"))

        # Section 4
        #
        # A variable number of sub-authority values
        #
        # Format: 4 byte little-endian unsigned longs
        # Human-Readable Section(s): 'S-1-5-21-1646645597-476216284-2431080400-80656'
        #                                   ^^ ^^^^^^^^^^ ^^^^^^^^^ ^^^^^^^^^^ ^^^^^
        # Bytes Section: \x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x10;\x01\x00
        #                                                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        for chunk in chunks:
            encoded_chunks.append(struct.pack("<L", chunk))
        return EncodedLdapBytes(b"".join(encoded_chunks), cls.name)

    @classmethod
    def query_encode(cls, data: Union[str, DecodedLdapString, EncodedLdapBytes]) -> EncodedQueryString:
        if not data:
            raise exceptions.AttributeEncodingError(f"Cannot encode blank {cls.name} {repr(data)}.")

        logger.debug(f"query encoding {cls.name} {data}")

        def _from_bytes(data):
            return EncodedQueryString(cls.decode(data), cls.name)

        def _from_str(data):
            return EncodedQueryString(data.__str__(), cls.name)

        @functools.singledispatch
        def _encode(data):
            raise exceptions.AttributeEncodingError(
                f"Cannot query-encode {cls.name} from data {data} of type {type(data)}."
            )

        @_encode.register
        def _str(data: str):
            return _from_str(data)

        @_encode.register
        def _decoded_str(data: DecodedLdapString):
            return _from_str(data)

        @_encode.register
        def _bytes(data: bytes):
            return _from_bytes(data)

        @_encode.register
        def _encoded_bytes(data: EncodedLdapBytes):
            return _from_bytes(data)

        return _encode(data)


@dataclasses.dataclass
class AttributeCodecs:
    logonHours = LogonHoursCodec
    objectSid = ObjectSidCodec
    objectGUID = ObjectGUIDCodec
    thumbnailPhoto = ThumbnailPhotoCodec
